@@ -1,27 +1,30 @@
 """Internal (private) Data Types Module."""
 
+import datetime
 import logging
 import re
 from decimal import Decimal
-from typing import Any, Dict, List, Match, Optional, Sequence, Tuple
+from typing import TYPE_CHECKING, Any, Dict, List, Match, Optional, Sequence, Tuple
 
-import pandas as pd  # type: ignore
-import pyarrow as pa  # type: ignore
-import pyarrow.parquet  # type: ignore
-import sqlalchemy  # type: ignore
-import sqlalchemy.dialects.mysql  # type: ignore
-import sqlalchemy.dialects.postgresql  # type: ignore
-import sqlalchemy_redshift.dialect  # type: ignore
-from sqlalchemy.sql.visitors import VisitableType  # type: ignore
+import numpy as np
+import pandas as pd
+import pyarrow as pa
+import pyarrow.parquet
+import sqlalchemy
+import sqlalchemy.dialects.mysql
+import sqlalchemy.dialects.postgresql
 
-from awswrangler import exceptions
+from awswrangler import _utils, exceptions
+
+if TYPE_CHECKING:
+    from sqlalchemy.sql.visitors import VisitableType
 
 _logger: logging.Logger = logging.getLogger(__name__)
 
 
 def athena2pyarrow(dtype: str) -> pa.DataType:  # pylint: disable=too-many-return-statements
     """Athena to PyArrow data types conversion."""
-    dtype = dtype.lower()
+    dtype = dtype.lower().replace(" ", "")
     if dtype == "tinyint":
         return pa.int8()
     if dtype == "smallint":
@@ -30,7 +33,7 @@ def athena2pyarrow(dtype: str) -> pa.DataType:  # pylint: disable=too-many-retur
         return pa.int32()
     if dtype == "bigint":
         return pa.int64()
-    if dtype == "float":
+    if dtype in ("float", "real"):
         return pa.float32()
     if dtype == "double":
         return pa.float64()
@@ -44,10 +47,16 @@ def athena2pyarrow(dtype: str) -> pa.DataType:  # pylint: disable=too-many-retur
         return pa.date32()
     if dtype in ("binary" or "varbinary"):
         return pa.binary()
-    if dtype.startswith("decimal"):
+    if dtype.startswith("decimal") is True:
         precision, scale = dtype.replace("decimal(", "").replace(")", "").split(sep=",")
         return pa.decimal128(precision=int(precision), scale=int(scale))
-    raise exceptions.UnsupportedType(f"Unsupported Athena type: {dtype}")  # pragma: no cover
+    if dtype.startswith("array") is True:
+        return pa.list_(value_type=athena2pyarrow(dtype=dtype[6:-1]), list_size=-1)
+    if dtype.startswith("struct") is True:
+        return pa.struct([(f.split(":", 1)[0], athena2pyarrow(f.split(":", 1)[1])) for f in dtype[7:-1].split(",")])
+    if dtype.startswith("map") is True:
+        return pa.map_(athena2pyarrow(dtype[4:-1].split(",", 1)[0]), athena2pyarrow(dtype[4:-1].split(",", 1)[1]))
+    raise exceptions.UnsupportedType(f"Unsupported Athena type: {dtype}")
 
 
 def athena2pandas(dtype: str) -> str:  # pylint: disable=too-many-branches,too-many-return-statements
@@ -61,7 +70,7 @@ def athena2pandas(dtype: str) -> str:  # pylint: disable=too-many-branches,too-m
         return "Int32"
     if dtype == "bigint":
         return "Int64"
-    if dtype == "float":
+    if dtype in ("float", "real"):
         return "float32"
     if dtype == "double":
         return "float64"
@@ -77,9 +86,7 @@ def athena2pandas(dtype: str) -> str:  # pylint: disable=too-many-branches,too-m
         return "decimal"
     if dtype in ("binary", "varbinary"):
         return "bytes"
-    if dtype == "array":  # pragma: no cover
-        return "list"
-    raise exceptions.UnsupportedType(f"Unsupported Athena type: {dtype}")  # pragma: no cover
+    raise exceptions.UnsupportedType(f"Unsupported Athena type: {dtype}")
 
 
 def athena2redshift(  # pylint: disable=too-many-branches,too-many-return-statements
@@ -87,13 +94,15 @@ def athena2redshift(  # pylint: disable=too-many-branches,too-many-return-statem
 ) -> str:
     """Athena to Redshift data types conversion."""
     dtype = dtype.lower()
+    if dtype == "tinyint":
+        return "SMALLINT"
     if dtype == "smallint":
         return "SMALLINT"
     if dtype in ("int", "integer"):
         return "INTEGER"
     if dtype == "bigint":
         return "BIGINT"
-    if dtype == "float":
+    if dtype in ("float", "real"):
         return "FLOAT4"
     if dtype == "double":
         return "FLOAT8"
@@ -107,7 +116,37 @@ def athena2redshift(  # pylint: disable=too-many-branches,too-many-return-statem
         return "DATE"
     if dtype.startswith("decimal"):
         return dtype.upper()
-    raise exceptions.UnsupportedType(f"Unsupported Athena type: {dtype}")  # pragma: no cover
+    raise exceptions.UnsupportedType(f"Unsupported Redshift type: {dtype}")
+
+
+def athena2quicksight(dtype: str) -> str:  # pylint: disable=too-many-branches,too-many-return-statements
+    """Athena to Quicksight data types conversion."""
+    dtype = dtype.lower()
+    if dtype == "tinyint":
+        return "INTEGER"
+    if dtype == "smallint":
+        return "INTEGER"
+    if dtype in ("int", "integer"):
+        return "INTEGER"
+    if dtype == "bigint":
+        return "INTEGER"
+    if dtype in ("float", "real"):
+        return "DECIMAL"
+    if dtype == "double":
+        return "DECIMAL"
+    if dtype in ("boolean", "bool"):
+        return "BOOLEAN"
+    if dtype in ("string", "char", "varchar"):
+        return "STRING"
+    if dtype == "timestamp":
+        return "DATETIME"
+    if dtype == "date":
+        return "DATETIME"
+    if dtype.startswith("decimal"):
+        return "DECIMAL"
+    if dtype == "binary":
+        return "BIT"
+    raise exceptions.UnsupportedType(f"Unsupported Athena type: {dtype}")
 
 
 def pyarrow2athena(dtype: pa.DataType) -> str:  # pylint: disable=too-many-branches,too-many-return-statements
@@ -143,21 +182,21 @@ def pyarrow2athena(dtype: pa.DataType) -> str:  # pylint: disable=too-many-branc
     if pa.types.is_list(dtype):
         return f"array<{pyarrow2athena(dtype=dtype.value_type)}>"
     if pa.types.is_struct(dtype):
-        return f"struct<{', '.join([f'{f.name}:{pyarrow2athena(dtype=f.type)}' for f in dtype])}>"
-    if pa.types.is_map(dtype):  # pragma: no cover
-        return f"map<{pyarrow2athena(dtype=dtype.key_type)},{pyarrow2athena(dtype=dtype.item_type)}>"
+        return f"struct<{','.join([f'{f.name}:{pyarrow2athena(dtype=f.type)}' for f in dtype])}>"
+    if pa.types.is_map(dtype):
+        return f"map<{pyarrow2athena(dtype=dtype.key_type)}, {pyarrow2athena(dtype=dtype.item_type)}>"
     if dtype == pa.null():
         raise exceptions.UndetectedType("We can not infer the data type from an entire null object column")
-    raise exceptions.UnsupportedType(f"Unsupported Pyarrow type: {dtype}")  # pragma: no cover
+    raise exceptions.UnsupportedType(f"Unsupported Pyarrow type: {dtype}")
 
 
 def pyarrow2pandas_extension(  # pylint: disable=too-many-branches,too-many-return-statements
-    dtype: pa.DataType
+    dtype: pa.DataType,
 ) -> Optional[pd.api.extensions.ExtensionDtype]:
     """Pyarrow to Pandas data types conversion."""
-    if pa.types.is_int8(dtype):  # pragma: no cover
+    if pa.types.is_int8(dtype):
         return pd.Int8Dtype()
-    if pa.types.is_int16(dtype):  # pragma: no cover
+    if pa.types.is_int16(dtype):
         return pd.Int16Dtype()
     if pa.types.is_int32(dtype):
         return pd.Int32Dtype()
@@ -172,7 +211,7 @@ def pyarrow2pandas_extension(  # pylint: disable=too-many-branches,too-many-retu
 
 def pyarrow2sqlalchemy(  # pylint: disable=too-many-branches,too-many-return-statements
     dtype: pa.DataType, db_type: str
-) -> Optional[VisitableType]:
+) -> Optional["VisitableType"]:
     """Pyarrow to Athena data types conversion."""
     if pa.types.is_int8(dtype):
         return sqlalchemy.types.SmallInteger
@@ -190,10 +229,12 @@ def pyarrow2sqlalchemy(  # pylint: disable=too-many-branches,too-many-return-sta
         if db_type == "postgresql":
             return sqlalchemy.dialects.postgresql.DOUBLE_PRECISION
         if db_type == "redshift":
+            import sqlalchemy_redshift.dialect  # pylint: disable=import-outside-toplevel
+
             return sqlalchemy_redshift.dialect.DOUBLE_PRECISION
         raise exceptions.InvalidDatabaseType(
             f"{db_type} is a invalid database type, please choose between postgresql, mysql and redshift."
-        )  # pragma: no cover
+        )
     if pa.types.is_boolean(dtype):
         return sqlalchemy.types.Boolean
     if pa.types.is_string(dtype):
@@ -205,22 +246,22 @@ def pyarrow2sqlalchemy(  # pylint: disable=too-many-branches,too-many-return-sta
             return sqlalchemy.types.VARCHAR(length=256)
         raise exceptions.InvalidDatabaseType(
             f"{db_type} is a invalid database type. " f"Please choose between postgresql, mysql and redshift."
-        )  # pragma: no cover
+        )
     if pa.types.is_timestamp(dtype):
         return sqlalchemy.types.DateTime
     if pa.types.is_date(dtype):
         return sqlalchemy.types.Date
     if pa.types.is_binary(dtype):
         if db_type == "redshift":
-            raise exceptions.UnsupportedType("Binary columns are not supported for Redshift.")  # pragma: no cover
+            raise exceptions.UnsupportedType("Binary columns are not supported for Redshift.")
         return sqlalchemy.types.Binary
     if pa.types.is_decimal(dtype):
         return sqlalchemy.types.Numeric(precision=dtype.precision, scale=dtype.scale)
     if pa.types.is_dictionary(dtype):
         return pyarrow2sqlalchemy(dtype=dtype.value_type, db_type=db_type)
-    if dtype == pa.null():  # pragma: no cover
+    if dtype == pa.null():
         return None
-    raise exceptions.UnsupportedType(f"Unsupported Pyarrow type: {dtype}")  # pragma: no cover
+    raise exceptions.UnsupportedType(f"Unsupported Pyarrow type: {dtype}")
 
 
 def pyarrow_types_from_pandas(
@@ -253,8 +294,20 @@ def pyarrow_types_from_pandas(
         _logger.debug("Inferring PyArrow type from column: %s", col)
         try:
             schema: pa.Schema = pa.Schema.from_pandas(df=df[[col]], preserve_index=False)
-        except pa.ArrowInvalid as ex:  # pragma: no cover
+        except pa.ArrowInvalid as ex:
             cols_dtypes[col] = process_not_inferred_dtype(ex)
+        except TypeError as ex:
+            msg = str(ex)
+            if " is required (got type " in msg:
+                raise TypeError(
+                    f"The {col} columns has a too generic data type ({df[col].dtype}) and seems "
+                    f"to have mixed data types ({msg}). "
+                    "Please, cast this columns with a more deterministic data type "
+                    f"(e.g. df['{col}'] = df['{col}'].astype('string')) or "
+                    "pass the column schema as argument for AWS Data Wrangler "
+                    f"(e.g. dtype={{'{col}': 'string'}}"
+                ) from ex
+            raise
         else:
             cols_dtypes[col] = schema.field(col).type
 
@@ -281,23 +334,23 @@ def process_not_inferred_dtype(ex: pa.ArrowInvalid) -> pa.DataType:
     """Infer data type from PyArrow inference exception."""
     ex_str = str(ex)
     _logger.debug("PyArrow was not able to infer data type:\n%s", ex_str)
-    match: Optional[Match] = re.search(
+    match: Optional[Match[str]] = re.search(
         pattern="Could not convert (.*) with type (.*): did not recognize "
         "Python value type when inferring an Arrow data type",
         string=ex_str,
     )
     if match is None:
-        raise ex  # pragma: no cover
+        raise ex
     groups: Optional[Sequence[str]] = match.groups()
     if groups is None:
-        raise ex  # pragma: no cover
+        raise ex
     if len(groups) != 2:
-        raise ex  # pragma: no cover
+        raise ex
     _logger.debug("groups: %s", groups)
     type_str: str = groups[1]
     if type_str == "UUID":
         return pa.string()
-    raise ex  # pragma: no cover
+    raise ex
 
 
 def process_not_inferred_array(ex: pa.ArrowInvalid, values: Any) -> pa.Array:
@@ -306,7 +359,7 @@ def process_not_inferred_array(ex: pa.ArrowInvalid, values: Any) -> pa.Array:
     if dtype == pa.string():
         array: pa.Array = pa.array(obj=[str(x) for x in values], type=dtype, safe=True)
     else:
-        raise ex  # pragma: no cover
+        raise ex
     return array
 
 
@@ -321,9 +374,21 @@ def athena_types_from_pandas(
     athena_columns_types: Dict[str, str] = {}
     for k, v in pa_columns_types.items():
         if v is None:
-            athena_columns_types[k] = casts[k]
+            athena_columns_types[k] = casts[k].replace(" ", "")
         else:
-            athena_columns_types[k] = pyarrow2athena(dtype=v)
+            try:
+                athena_columns_types[k] = pyarrow2athena(dtype=v)
+            except exceptions.UndetectedType as ex:
+                raise exceptions.UndetectedType(
+                    "Impossible to infer the equivalent Athena data type "
+                    f"for the {k} column. "
+                    "It is completely empty (only null values) "
+                    f"and has a too generic data type ({df[k].dtype}). "
+                    "Please, cast this columns with a more deterministic data type "
+                    f"(e.g. df['{k}'] = df['{k}'].astype('string')) or "
+                    "pass the column schema as argument for AWS Data Wrangler "
+                    f"(e.g. dtype={{'{k}': 'string'}}"
+                ) from ex
     _logger.debug("athena_columns_types: %s", athena_columns_types)
     return athena_columns_types
 
@@ -341,12 +406,12 @@ def athena_types_from_pandas_partitioned(
         df=df, index=index, dtype=dtype, index_left=index_left
     )
     columns_types: Dict[str, str] = {}
+    for col, typ in athena_columns_types.items():
+        if col not in partitions:
+            columns_types[col] = typ
     partitions_types: Dict[str, str] = {}
-    for k, v in athena_columns_types.items():
-        if k in partitions:
-            partitions_types[k] = v
-        else:
-            columns_types[k] = v
+    for par in partitions:
+        partitions_types[par] = athena_columns_types[par]
     return columns_types, partitions_types
 
 
@@ -362,7 +427,7 @@ def pyarrow_schema_from_pandas(
     )
     for k, v in casts.items():
         if (k in df.columns) and (k not in ignore):
-            columns_types[k] = athena2pyarrow(v)
+            columns_types[k] = athena2pyarrow(dtype=v)
     columns_types = {k: v for k, v in columns_types.items() if v is not None}
     _logger.debug("columns_types: %s", columns_types)
     return pa.schema(fields=columns_types)
@@ -381,51 +446,80 @@ def athena_types_from_pyarrow_schema(
     return columns_types, partitions_types
 
 
-def athena_partitions_from_pyarrow_partitions(
-    path: str, partitions: pyarrow.parquet.ParquetPartitions
-) -> Dict[str, List[str]]:
-    """Extract the related Athena partitions values from any PyArrow Partitions."""
-    path = path if path[-1] == "/" else f"{path}/"
-    partitions_values: Dict[str, List[str]] = {}
-    names: List[str] = [p.name for p in partitions]
-    for values in zip(*[p.keys for p in partitions]):
-        suffix: str = "/".join([f"{n}={v}" for n, v in zip(names, values)])
-        suffix = suffix if suffix[-1] == "/" else f"{suffix}/"
-        partitions_values[f"{path}{suffix}"] = list(values)
-    return partitions_values
-
-
 def cast_pandas_with_athena_types(df: pd.DataFrame, dtype: Dict[str, str]) -> pd.DataFrame:
     """Cast columns in a Pandas DataFrame."""
+    mutability_ensured: bool = False
     for col, athena_type in dtype.items():
-        if col in df.columns:
-            pandas_type: str = athena2pandas(dtype=athena_type)
-            if pandas_type == "datetime64":
-                df[col] = pd.to_datetime(df[col])
-            elif pandas_type == "date":
-                df[col] = pd.to_datetime(df[col]).dt.date.replace(to_replace={pd.NaT: None})
-            elif pandas_type == "bytes":
-                df[col] = df[col].astype("string").str.encode(encoding="utf-8").replace(to_replace={pd.NA: None})
-            elif pandas_type == "decimal":
-                df[col] = (
-                    df[col]
-                    .astype("string")
-                    .apply(lambda x: Decimal(str(x)) if str(x) not in ("", "none", " ", "<NA>") else None)
-                )
-            else:
-                df[col] = df[col].astype(pandas_type)
+        if (
+            (col in df.columns)
+            and (athena_type.startswith("array") is False)
+            and (athena_type.startswith("struct") is False)
+            and (athena_type.startswith("map") is False)
+        ):
+            desired_type: str = athena2pandas(dtype=athena_type)
+            current_type: str = _normalize_pandas_dtype_name(dtype=str(df[col].dtypes))
+            if desired_type != current_type:  # Needs conversion
+                _logger.debug("current_type: %s -> desired_type: %s", current_type, desired_type)
+                if mutability_ensured is False:
+                    df = _utils.ensure_df_is_mutable(df=df)
+                    mutability_ensured = True
+                _cast_pandas_column(df=df, col=col, current_type=current_type, desired_type=desired_type)
+    return df
+
+
+def _normalize_pandas_dtype_name(dtype: str) -> str:
+    if dtype.startswith("datetime64") is True:
+        return "datetime64"
+    if dtype.startswith("decimal") is True:
+        return "decimal"
+    return dtype
+
+
+def _cast2date(value: Any) -> Any:
+    if isinstance(value, float) and (np.isnan(value) or np.isinf(value)):
+        return None
+    if pd.isna(value) or value is None:
+        return None
+    if isinstance(value, datetime.date):
+        return value
+    return pd.to_datetime(value).date()
+
+
+def _cast_pandas_column(df: pd.DataFrame, col: str, current_type: str, desired_type: str) -> pd.DataFrame:
+    if desired_type == "datetime64":
+        df[col] = pd.to_datetime(df[col])
+    elif desired_type == "date":
+        df[col] = df[col].apply(lambda x: _cast2date(value=x)).replace(to_replace={pd.NaT: None})
+    elif desired_type == "bytes":
+        df[col] = df[col].astype("string").str.encode(encoding="utf-8").replace(to_replace={pd.NA: None})
+    elif desired_type == "decimal":
+        # First cast to string
+        df = _cast_pandas_column(df=df, col=col, current_type=current_type, desired_type="string")
+        # Then cast to decimal
+        df[col] = df[col].apply(lambda x: Decimal(str(x)) if str(x) not in ("", "none", "None", " ", "<NA>") else None)
+    else:
+        try:
+            df[col] = df[col].astype(desired_type)
+        except TypeError as ex:
+            if "object cannot be converted to an IntegerDtype" not in str(ex):
+                raise ex
+            df[col] = (
+                df[col]
+                .apply(lambda x: int(x) if str(x) not in ("", "none", "None", " ", "<NA>") else None)
+                .astype(desired_type)
+            )
     return df
 
 
 def sqlalchemy_types_from_pandas(
-    df: pd.DataFrame, db_type: str, dtype: Optional[Dict[str, VisitableType]] = None
-) -> Dict[str, VisitableType]:
+    df: pd.DataFrame, db_type: str, dtype: Optional[Dict[str, "VisitableType"]] = None
+) -> Dict[str, "VisitableType"]:
     """Extract the related SQLAlchemy data types from any Pandas DataFrame."""
-    casts: Dict[str, VisitableType] = dtype if dtype is not None else {}
+    casts: Dict[str, "VisitableType"] = dtype if dtype is not None else {}
     pa_columns_types: Dict[str, Optional[pa.DataType]] = pyarrow_types_from_pandas(
         df=df, index=False, ignore_cols=list(casts.keys())
     )
-    sqlalchemy_columns_types: Dict[str, VisitableType] = {}
+    sqlalchemy_columns_types: Dict[str, "VisitableType"] = {}
     for k, v in pa_columns_types.items():
         if v is None:
             sqlalchemy_columns_types[k] = casts[k]
